#!/usr/bin/env python

# PWHN Project
# Copyright (C) 2006 Vitus Lorenz-Meyer, University of Texas @ El Paso
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA	02111-1307	USA

"""
PSMRWAE script: Planetary-Scale Map-Reduce (operations) With Arbitrary Executables
"""

version = 0.8
import getopt, os, sys
from cStringIO import StringIO
script,t = os.path.splitext(sys.argv[0])
t,scriptname = os.path.split(script)

def main():
	
	if len(sys.argv) == 1:
		print "No command given."
		usage()
		sys.exit(1)		 

	
	options = 'F:hi:D:Vu:g:e:sql\
				P:ra:o:yt:pn'
	long_opts = ['file=','help','init=','dir=','version','user=','agg=','eval=','nosync','quiet','synconly', \
			'max-procs=','randomize','args=','output=','noprompt','timeout=','plain','no-exec']
	
	try:
		opts,rest = getopt.getopt(sys.argv[1:], options,long_opts)
	except getopt.GetoptError, e:
		print "Error Parsing Options: %s" % str(e)
		usage()
		sys.exit(1)

	varsset=0
	for o,a in opts:
		if o in ['--file','-F']:
			file = open(a, "r")
			exec(file)
			file.close()
			varsset=1

	if(not varsset):
		syncdir = 	''
		initfile = 	''
		aggfile = 	''
		evalfile = 	''
		outdir =	'/tmp/'+scriptname+'-out'
		username = 	''
		nosync = 	0
		synconly = 	0
		quiet = 	0
		vxargs = 	[]
	
	for o,a in opts:
		if o in ['--file','-F']:
			continue
		elif o in ['--nosync','-s']:
			nosync=1
		elif o in ['--synconly','-l']:
			synconly=1
		elif o in ['--quiet','-q']:
			quiet=1
			vxargs.append('-p')
			vxargs.append('-y')
		elif o in ['--version','-V']:
			print "%s version %s" % (scriptname,version)
			print "Copyright (c) 2006 Vitus Lorenz-Meyer (vdlorenz@utep.edu)"
			print "Freely distributed under GNU LGPL License"
			sys.exit(1)
		elif o in ['-h','--help']:
			usage()
			sys.exit(1)
		elif o in ['-u','--user']:
			username = "%s@" % a
		elif o in ['-g','--agg']:
			if (syncdir and not a.startswith(syncdir)):
				aggfile = os.path.join(syncdir,a)
			else:
				aggfile = a
			if not os.path.exists(aggfile) or not os.path.isfile(aggfile):
				  print "argument file %s does not exists or is not a plain file" % aggfile
				  sys.exit(3)
		elif o in ['-i','--init']:
			if (syncdir and (not(a.startswith(syncdir)))):
				initfile = os.path.join(syncdir,a)
			else:
				initfile = a
			if not os.path.exists(initfile) or not os.path.isfile(initfile):
				  print "argument file %s does not exists or is not a plain file" % initfile
				  sys.exit(3)
		elif o in ['-e','--eval']:
			if (syncdir and not a.startswith(syncdir)):
				evalfile = os.path.join(syncdir,a)
			else:
				evalfile = a
			if not os.path.exists(evalfile) or not os.path.isfile(evalfile):
				  print "argument file %s does not exists or is not a plain file" % evalfile
				  sys.exit(3)
		elif o in ['-D','--dir']:
			if os.path.isdir(a):
				syncdir = a
				if syncdir.endswith(os.sep):
					syncdir,t = os.path.split(syncdir)
			else:
				print "Directory '%s' does not exist" % a
				sys.exit(3)
		elif o in ['-P','-r','-a','-o','-y','-t','-p','-n', \
				'max-procs=','randomize','args=','output=','noprompt','timeout=','plain','no-exec']:
			if o in ['-o', '--output']:
				outdir=a
			else:
				vxargs.append(' '.join([o,a]))
			
		else:
			print 'Unknown options: %s %s' % (o,a)
			usage()
			sys.exit(1)
	
	
	if (not(initfile) and not(synconly)):
		print "No Initfile given"
		usage()
		sys.exit(3)
				
	vxargs.append(' '.join(['-o',outdir]))
	#rsync first with the help of vxargs
	if (syncdir):
		rsyncfile = syncdir
		t,syncdir_name = os.path.split(syncdir)
	else:
		rsyncfile = initfile
		
	sync = 'python vxargs.py %s rsync -arz %s %s{}:~/%s/' % (' '.join(vxargs),rsyncfile,username,scriptname)
	vxargs.append("-y")
	t,initfile_name = os.path.split(initfile)
	if (syncdir):
		initfile_name = os.path.join(syncdir_name,initfile_name)

	ssh = 'python vxargs.py %s ssh %s{} "~/%s/%s"' % (' '.join(vxargs),username,scriptname,initfile_name)

	#print "Execing %s" %sync
	if not nosync:
		if (not quiet):
			print "---- vxargs rsync ----"
			os.system(sync)
		else:
			in_pipe,out_pipe = os.popen2(sync)
			in_pipe.close()
			out_pipe.readlines()
			out_pipe.close()
				
	#and kick off remote execution with the help of vxargs
	#print "Execing %s" %ssh
	if not synconly:
		if (not quiet):
			print "---- vxargs ssh ----"
			os.system(ssh)
		else:
			in_pipe,out_pipe = os.popen2(ssh)
			in_pipe.close()
			out_pipe.readlines()
			out_pipe.close()
		
		#read results from outdir
		results = []
		for fn in os.listdir(outdir):
			if os.path.isfile(os.path.join(outdir,fn)) and fn.endswith('.out') :
				f = open(os.path.join(outdir,fn),'r')
				results.append(''.join(f.readlines()))
	
	
		if (not quiet):
			print "---- Results ----"
		# if theres no agg nor eval, just output everything
		if (not(aggfile) and not(evalfile)):
			print "\n".join(results)
			sys.exit(0)
	
		#and run through aggregator script, if defined
		if (aggfile):
			res = results.pop(0)
			for r in results:
				in_pipe,out_pipe = os.popen2(aggfile)
				#write length of first input to file, without trailing \n
				in_pipe.write("%s\n" % len(res))
				#and input
				in_pipe.write("%s\n" % res)
				#write length of 2nd input to file, without trailing \n
				in_pipe.write("%s\n" % len(r))
				#and input
				in_pipe.write("%s\n" % r)
				#wait for it to finish
				in_pipe.close()
				#now read
				res2 = ''.join(out_pipe.readlines())
				out_pipe.close()
				if (not quiet):
					print "Agg(%s + %s) = %s" % (res,r,res2)
				res = res2
		else:
			file_str = StringIO()
			#evalfile now expects all results seperated by newlines and length
			for r in results:
				file_str.write("%s\n" % len(r))
				file_str.write("%s\n" % r)
			res = file_str.getvalue()
		
		
		#run through eval script, if any
		if (evalfile):
			in_pipe,out_pipe = os.popen2(evalfile)
			in_pipe.write(res)
			in_pipe.close()
			#now read
			evald = ''.join(out_pipe.readlines())
			out_pipe.close()
			if (not quiet):
				print "Eval(%s) = %s" % (res,evald)
			else:
				print evald
		else:
			#just print the result before evaling
			print res


def handler(signum, frame_unused):
	if signum==signal.SIGINT:
		sys.exit(1)

def usage():
	print """\
Usage:

NAME
  """ + scriptname + """ - simple map-reduce script with arbitrary executables.
   
DESCRIPTION
  Syncs a file or directory with a remote server and executes init script remotely using vxargs.
  Most of the options are passed directly to vxargs.
  For vxargs to work the shell has to know where to find an ssh-agent that has the right identity added.
  
SYNOPSIS
  """ + scriptname + """ [OPTIONS] -a <file> -i <file>

OPTIONS
  --file=filename, -f file
	Sources a file that specifies all options.
	Certain options may be overriden on the commandline.
	File is in Python syntax and should look like this (default settings):
		syncdir = 	''
		initfile = 	''
		aggfile = 	''
		evalfile = 	''
		outdir =	'/tmp/'+scriptname+'-out'
		username = 	''
		nosync = 	0
		synconly = 	0
		quiet = 	0
		vxargs = []		#list of single strings to give to vxargs, ex: [ "-a ips.txt" ]
		
  --nosync, -s
  	If every server is already up to date, syncing can be switched of by specifying --no-sync to speed up execution.

  --synconly, -l
  	To only sync, on the other hand, without executing, use synconly.
  	
  --quiet, -q
  	Supresses any output including output from vxargs, the only thing written to stdout is the last result.
  	Implies -p and -y for vxargs.
  	
  --help
	Print a summary of the options	 and exit.

  --init=filename, -i file
	Specifies the name of the initializer script to run remotely.
	If <dir> is not specified only this file will be synced.
	
  --agg=filename, -g file
	Specifies the name of the aggregator script.
	If <dir> is specified this file will be assumed to be in <dir>.

  --eval=filename, -e file
	Specifies the name of the evaluator script.
	If <dir> is specified this file will be assumed to be in <dir>.

  --dir=dirname, -D dir
	The directory to syncronise. If specified all files have to be in that directory.
 
  --user=username, -u username
	The username to use for remote login, if different from current logged in user.

  --version
	Display current version and copyright information.


 Options passed to vxargs
  --max-procs=max-procs, -P max-procs
	Run up to max-procs processes at a time; the default is 30.

  --randomize, -r
	Randomize the host list before all execution.
	
  --args=filename, -a filename
	The arguments file.
	
  --output=outdir, -o outdir
	output directory for stdout and stderr files
	The default value is specified by the environment variable VXARGS_OUTDIR.
	If it is unspecified, both stdout and stderr will be redirected
	to a temp directory.
	Note that if the directory existed before execution, everything
	inside will be wiped without warning.
	
  --timeout=timeout, -t timeout
	The maximal time in second for each command to execute. timeout=0
	means infinite.	 0 (i.e. infinite) is the default value. When the time is up,
	vxargs will send signal SIGINT to the process. If the process does not
	stop after 2 seconds, vxargs will send SIGTERM signal, and send SIGKILL
	if it still keeps running after 3 seconds.

  --noprompt, -y
	Wipe out the outdir without confirmation.

  --no-exec, -n
	Print the commands that would be executed, but do not execute them.

  --plain, -p
	Don't use curses-based output, but plain output to stdout
	instead. It will be less exciting, but will do the same job
	effectively.
	By default, """ + scriptname + """ uses the curses-based output from vxargs.

SCRIPTS

 Initialiser
 	The init executable is run remotely.
 	It has to be an executable file, but could also be a script that is executable and has a working shebang.
 	It is expected to write its output to stdout.
 	An output could look like this, for example:
 		(1,3.0,2.0)
	If this is the only script given, the output of every run will be printed to stdout seperated by newlines.
 	
 Aggregator
 	The agg executable is run locally.
 	It has to be an executable file, but could also be a script that is executable and has a working shebang.
	It will receive the outputs of 2 init runs on stdin and is expected to write one aggregate to stdout.
	The inputs will both start with the exact length in bytes, written in ASCII on a line by itself.
	Then the input followed by a newline which is not counted as part of the input.
	The input could look like this, for example:
		11\\n
		(1,3.0,2.0)\\n
		12\\n
		(1,1.65,0.8)\\n
	Then the output could look like this:
		(2,4.65,2.8)
	If no aggregator is given, the input to the evaluator will be all outputs of the init script, seperated by the lengths in ASCII on lines by
	themselves like the input to the aggregator, except possibly consisting of more than two records.
		
 Evaluator
	The eval executable is run locally.
	It has to be an executable file, but could also be a script that is executable and has a working shebang.
	It will receive the last aggregate value on stdin and is expected to deliver the final result to stdout.
	Continuing the above example, the final result for the input:
		(2,4.65,2.8)
	could be:
		(2.325,1.4)
	If no evaluator is given, the output of the last aggregator will be written to stdout.
	
		
EXAMPLE
	Suppose the following conditions hold (the files have the stated contents).
	$ cat ips.txt
	192.41.135.218
	#planetlab1.csg.unizh.ch
	129.108.202.10
	#planetlab1.utep.edu
	131.175.17.10
	#planetlab2.elet.polimi.it

	$ cat scripts/init.php
	#!/usr/bin/php -f
	<?php
		$u = `uptime`;
		$load = trim(substr($u,strrpos($u,':')+1));
		$vals = explode( (strpos($load,',')>0)?',':' ',$load);
		printf("(1,%s,%s,%s)",$vals[0],$vals[1],$vals[2]);
	?>
	
	$ cat scripts/agg.php
	#!/usr/bin/php -f
	<?php
		fscanf(STDIN, "%d\\n", $l1);
		$first = fread(STDIN,$l1);
		fread(STDIN,1);
		fscanf(STDIN, "%d\\n", $l2);
		$second = fread(STDIN,$l2);
		$v1 = explode(',',substr($first,1,strlen($first)-2));
		$v2 = explode(',',substr($second,1,strlen($second)-2));
		printf("(%u,%f,%f,%f)",
						(round($v1[0]+$v2[0])),
						($v1[1]+$v2[1]),
						($v1[2]+$v2[2]),
						($v1[3]+$v2[3])
		);
	?>
	
	$ cat scripts/eval.php
	#!/usr/bin/php -f
	<?php
		$in = trim(stream_get_contents(STDIN));
		$v = explode(',',substr($in,1,strlen($in)-2));
		printf("(%.4f,%.4f,%.4f)",
						($v[1]/$v[0]),
						($v[2]/$v[0]),
						($v[3]/$v[0])
		);
	?>

	Then executing
		./""" + scriptname + """.py -D scripts/ -i init.php -g agg.php -e eval.php -a ips.txt
	will output something similar to
		exit code 0: 3 job(s)
		total number of jobs: 3
		exit code 0: 3 job(s)
		total number of jobs: 3
		Agg((1,2.57, 2.64, 2.64) + (1,1.73, 1.84, 1.84)) = (2,4.300000,4.480000,4.480000)
		Agg((2,4.300000,4.480000,4.480000) + (1,1.55, 1.40, 1.42)) = (3,5.850000,5.880000,5.900000)
		Eval((3,5.850000,5.880000,5.900000)) = (1.9500,1.9600,1.9667)
		
""" 

if __name__=='__main__':
	main()
